@@ -20,26 +20,7 @@ repeatedly dealing with errors as appropriate.  Finally
 :meth:`~Backup.finish` cleans up committing or rolling back and
 releasing locks.
 
-Here is an example usage using the **with** statement to ensure
-:meth:`~Backup.finish` is called::
-
-  # copies source.main into db
-  with db.backup("main", source, "main") as b:
-      while not b.done:
-          b.step(100)
-          print(b.remaining, b.pagecount, "\r", flush = True)
-
-If you are not using **with** then you'll need to ensure
-:meth:`~Backup.finish` is called::
-
-  # copies source.main into db
-  b=db.backup("main", source, "main")
-  try:
-      while not b.done:
-          b.step(100)
-          print(b.remaining, b.pagecount, "\r", flush = True)
-  finally:
-      b.finish()
+See the :ref:`example <example_backup>`.
 
 Important details
 =================
@@ -49,6 +30,9 @@ round trip via SQL.  All pages are copied including free ones.
 
 The destination database is locked during the copy.  You will get a
 :exc:`ThreadingViolationError` if you attempt to use it.
+
+The source database can change during the backup.  SQLite will
+come back and copy those changes too until the backup is complete.
 */
 
 /* we love us some macros */
@@ -119,13 +103,12 @@ APSWBackup_close_internal(APSWBackup *self, int force)
       break;
     case 2:
     {
-      PyObject *etype, *eval, *etb;
-      PyErr_Fetch(&etype, &eval, &etb);
+      PY_ERR_FETCH(exc_save);
 
       SET_EXC(res, self->dest->db);
       apsw_write_unraisable(NULL);
 
-      PyErr_Restore(etype, eval, etb);
+      PY_ERR_RESTORE(exc_save);
       break;
     }
     }
@@ -164,21 +147,18 @@ APSWBackup_dealloc(APSWBackup *self)
   backup object is :meth:`finished <Backup.finish>`.
 
   :param npages: How many pages to copy. If the parameter is omitted
-     or negative then all remaining pages are copied. The default page
-     size is 4096 bytes (4kb) which can be viewed, or changed before database
-     creation using a `pragma
-     <https://www.sqlite.org/pragma.html#pragma_page_size>`_.
+     or negative then all remaining pages are copied.
 
   This method may throw a :exc:`BusyError` or :exc:`LockedError` if
   unable to lock the source database.  You can catch those and try
   again.
 
-  :returns: True if this copied the last remaining outstanding pages, else false.  This is the same value as :attr:`~Backup.done`
+  :returns: True if this copied the last remaining outstanding pages, else False.  This is the same value as :attr:`~Backup.done`
 
   -* sqlite3_backup_step
 */
 static PyObject *
-APSWBackup_step(APSWBackup *self, PyObject *args, PyObject *kwds)
+APSWBackup_step(APSWBackup *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int npages = -1, res;
 
@@ -186,10 +166,10 @@ APSWBackup_step(APSWBackup *self, PyObject *args, PyObject *kwds)
   CHECK_BACKUP_CLOSED(NULL);
 
   {
-    static char *kwlist[] = {"npages", NULL};
     Backup_step_CHECK;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|i:" Backup_step_USAGE, kwlist, &npages))
-      return NULL;
+    ARG_PROLOG(1, Backup_step_KWNAMES);
+    ARG_OPTIONAL ARG_int(npages);
+    ARG_EPILOG(NULL, Backup_step_USAGE, );
   }
   PYSQLITE_BACKUP_CALL(res = sqlite3_backup_step(self->backup, npages));
 
@@ -211,7 +191,7 @@ APSWBackup_step(APSWBackup *self, PyObject *args, PyObject *kwds)
 
   if (res)
   {
-    SET_EXC(res, NULL); /* ::TODO:: will likely have message on dest->db */
+    SET_EXC(res, NULL);
     return NULL;
   }
 
@@ -248,15 +228,13 @@ APSWBackup_finish(APSWBackup *self)
 /** .. method:: close(force: bool = False) -> None
 
   Does the same thing as :meth:`~Backup.finish`.  This extra api is
-  provided to give the same api as other APSW objects such as
-  :meth:`Connection.close`, :meth:`Blob.close` and
-  :meth:`Cursor.close`.  It is safe to call this method multiple
-  times.
+  provided to give the same api as other APSW objects and files.
+  It is safe to call this method multiple  times.
 
   :param force: If true then any exceptions are ignored.
 */
 static PyObject *
-APSWBackup_close(APSWBackup *self, PyObject *args, PyObject *kwds)
+APSWBackup_close(APSWBackup *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int force = 0, setexc;
 
@@ -267,11 +245,10 @@ APSWBackup_close(APSWBackup *self, PyObject *args, PyObject *kwds)
     Py_RETURN_NONE; /* already closed */
 
   {
-    static char *kwlist[] = {"force", NULL};
     Backup_close_CHECK;
-    argcheck_bool_param force_param = {&force, Backup_close_force_MSG};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O&:" Backup_close_USAGE, kwlist, argcheck_bool, &force_param))
-      return NULL;
+    ARG_PROLOG(1, Backup_close_KWNAMES);
+    ARG_OPTIONAL ARG_bool(force);
+    ARG_EPILOG(NULL, Backup_close_USAGE, );
   }
   setexc = APSWBackup_close_internal(self, force);
   if (setexc)
@@ -296,7 +273,7 @@ APSWBackup_get_remaining(APSWBackup *self, void *Py_UNUSED(ignored))
   return PyLong_FromLong(self->backup ? sqlite3_backup_remaining(self->backup) : 0);
 }
 
-/** .. attribute:: pagecount
+/** .. attribute:: page_count
   :type: int
 
   Read only. How many pages were in the source database after the last
@@ -307,7 +284,7 @@ APSWBackup_get_remaining(APSWBackup *self, void *Py_UNUSED(ignored))
   -* sqlite3_backup_pagecount
 */
 static PyObject *
-APSWBackup_get_pagecount(APSWBackup *self, void *Py_UNUSED(ignored))
+APSWBackup_get_page_count(APSWBackup *self, void *Py_UNUSED(ignored))
 {
   CHECK_USE(NULL);
   return PyLong_FromLong(self->backup ? sqlite3_backup_pagecount(self->backup) : 0);
@@ -316,7 +293,7 @@ APSWBackup_get_pagecount(APSWBackup *self, void *Py_UNUSED(ignored))
 /** .. method:: __enter__() -> Backup
 
   You can use the backup object as a `context manager
-  <https://docs.python.org/reference/datamodel.html#with-statement-context-managers>`_
+  <https://docs.python.org/3/reference/datamodel.html#with-statement-context-managers>`_
   as defined in :pep:`0343`.  The :meth:`~Backup.__exit__` method ensures that backup
   is :meth:`finished <Backup.finish>`.
 */
@@ -335,17 +312,19 @@ APSWBackup_enter(APSWBackup *self)
   that the copy is :meth:`finished <Backup.finish>`.
 */
 static PyObject *
-APSWBackup_exit(APSWBackup *self, PyObject *args, PyObject *kwds)
+APSWBackup_exit(APSWBackup *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   PyObject *etype, *evalue, *etraceback;
   int setexc;
 
   CHECK_USE(NULL);
   {
-    static char *kwlist[] = {"etype", "evalue", "etraceback", NULL};
     Backup_exit_CHECK;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "OOO:" Backup_exit_USAGE, kwlist, &etype, &evalue, &etraceback))
-      return NULL;
+    ARG_PROLOG(3, Backup_exit_KWNAMES);
+    ARG_MANDATORY ARG_pyobject(etype);
+    ARG_MANDATORY ARG_pyobject(evalue);
+    ARG_MANDATORY ARG_pyobject(etraceback);
+    ARG_EPILOG(NULL, Backup_exit_USAGE, );
   }
   /* If already closed then we are fine - CHECK_BACKUP_CLOSED not needed*/
   if (!self->backup)
@@ -367,6 +346,15 @@ APSWBackup_exit(APSWBackup *self, PyObject *args, PyObject *kwds)
   Py_RETURN_FALSE;
 }
 
+static PyObject *
+APSWBackup_tp_str(APSWBackup *self)
+{
+  return PyUnicode_FromFormat("<apsw.Backup object from %S to %S at %p>",
+                              self->source ? (PyObject *)self->source : apst.closed,
+                              self->dest ? (PyObject *)self->dest : apst.closed,
+                              self);
+}
+
 /** .. attribute:: done
   :type: bool
 
@@ -380,19 +368,22 @@ static PyMemberDef backup_members[] = {
 static PyGetSetDef backup_getset[] = {
     /* name getter setter doc closure */
     {"remaining", (getter)APSWBackup_get_remaining, NULL, Backup_remaining_DOC, NULL},
-    {"pagecount", (getter)APSWBackup_get_pagecount, NULL, Backup_pagecount_DOC, NULL},
+    {"page_count", (getter)APSWBackup_get_page_count, NULL, Backup_page_count_DOC, NULL},
+#ifndef APSW_OMIT_OLD_NAMES
+    {Backup_page_count_OLDNAME, (getter)APSWBackup_get_page_count, NULL, Backup_page_count_OLDDOC, NULL},
+#endif
     {0, 0, 0, 0, 0}};
 
 static PyMethodDef backup_methods[] = {
     {"__enter__", (PyCFunction)APSWBackup_enter, METH_NOARGS,
      Backup_enter_DOC},
-    {"__exit__", (PyCFunction)APSWBackup_exit, METH_VARARGS | METH_KEYWORDS,
+    {"__exit__", (PyCFunction)APSWBackup_exit, METH_FASTCALL | METH_KEYWORDS,
      Backup_exit_DOC},
-    {"step", (PyCFunction)APSWBackup_step, METH_VARARGS | METH_KEYWORDS,
+    {"step", (PyCFunction)APSWBackup_step, METH_FASTCALL | METH_KEYWORDS,
      Backup_step_DOC},
     {"finish", (PyCFunction)APSWBackup_finish, METH_NOARGS,
      Backup_finish_DOC},
-    {"close", (PyCFunction)APSWBackup_close, METH_VARARGS | METH_KEYWORDS,
+    {"close", (PyCFunction)APSWBackup_close, METH_FASTCALL | METH_KEYWORDS,
      Backup_close_DOC},
     {0, 0, 0, 0}};
 
@@ -408,4 +399,5 @@ static PyTypeObject APSWBackupType =
         .tp_methods = backup_methods,
         .tp_members = backup_members,
         .tp_getset = backup_getset,
+        .tp_str = (reprfunc)APSWBackup_tp_str,
 };
