@@ -12,7 +12,8 @@ Blob Input/Output
 
 A `blob <https://en.wikipedia.org/wiki/Binary_large_object>`_ is a
 SQLite `datatype <https://sqlite.org/datatype3.html>`_ representing
-a sequence of bytes.  It can be zero or more bytes in size.
+a sequence of bytes.  It can be zero or more bytes in size.  Blobs
+cannot be resized, but you can read and overwrite parts of them.
 
 SQLite blobs have an absolute maximum size of 2GB and a `default
 maximum size <https://sqlite.org/c3ref/c_limit_attached.html>`_ of
@@ -21,6 +22,7 @@ maximum size <https://sqlite.org/c3ref/c_limit_attached.html>`_ of
 An alternate approach to using blobs is to store the data in files and
 store the filename in the database.  Doing so loses the `ACID
 <https://sqlite.org/transactional.html>`_ properties of SQLite.
+There are `benchmarks <https://www.sqlite.org/fasterthanfs.html>`__.
 
 */
 
@@ -28,28 +30,14 @@ store the filename in the database.  Doing so loses the `ACID
 
 /** .. class:: zeroblob
 
-  If you want to insert a blob into a row, you previously needed to
-  supply the entire blob in one go.  To read just one byte also
-  required retrieving the blob in its entirety. For example to insert
-  a 100MB file you would have done::
+  If you want to insert a blob into a row, you need to
+  supply the entire blob in one go.  Using this class or
+  `function <https://www.sqlite.org/lang_corefunc.html#zeroblob>`__
+  allocates the space in the database filling it with zeroes.
 
-     largedata=open("largefile", "rb").read()
-     cur.execute("insert into foo values(?)", (largedata,))
-
-  SQLite 3.5 allowed for incremental Blob I/O so you can read and
-  write blobs in small amounts.  You cannot change the size of a blob
-  so you need to reserve space which you do through zeroblob which
-  creates a blob of the specified size but full of zero bytes.  For
-  example you would reserve space for your 100MB one of these two
-  ways::
-
-    cur.execute("insert into foo values(zeroblob(100000000))")
-    cur.execute("insert into foo values(?),
-                 (apsw.zeroblob(100000000),))
-
-  This class is used for the second way.  Once a blob exists in the
-  database, you then use the :class:`Blob` class to read and write its
-  contents.
+  You can then overwrite parts in smaller chunks, without having
+  to do it all at once.  The :ref:`example <example_blob_io>` shows
+  how to use it.
 */
 
 /* ZeroBlobBind is defined in apsw.c because of forward references */
@@ -60,7 +48,10 @@ ZeroBlobBind_new(PyTypeObject *type, PyObject *Py_UNUSED(args), PyObject *Py_UNU
   ZeroBlobBind *self;
   self = (ZeroBlobBind *)type->tp_alloc(type, 0);
   if (self)
+  {
     self->blobsize = 0;
+    self->init_was_called = 0;
+  }
   return (PyObject *)self;
 }
 
@@ -69,15 +60,17 @@ ZeroBlobBind_new(PyTypeObject *type, PyObject *Py_UNUSED(args), PyObject *Py_UNU
   :param size: Number of zeroed bytes to create
 */
 static int
-ZeroBlobBind_init(ZeroBlobBind *self, PyObject *args, PyObject *kwds)
+ZeroBlobBind_init(ZeroBlobBind *self, PyObject *args, PyObject *kwargs)
 {
   long long size;
 
   {
-    static char *kwlist[] = {"size", NULL};
     Zeroblob_init_CHECK;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "L:" Zeroblob_init_USAGE, kwlist, &size))
-      return -1;
+    PREVENT_INIT_MULTIPLE_CALLS;
+    ARG_CONVERT_VARARGS_TO_FASTCALL;
+    ARG_PROLOG(1, Zeroblob_init_KWNAMES);
+    ARG_MANDATORY ARG_int64(size);
+    ARG_EPILOG(-1, Zeroblob_init_USAGE, Py_XDECREF(fast_kwnames));
   }
   if (size < 0)
   {
@@ -99,6 +92,14 @@ ZeroBlobBind_len(ZeroBlobBind *self)
   return PyLong_FromLong(self->blobsize);
 }
 
+static PyObject *
+ZeroBlobBind_tp_str(ZeroBlobBind *self)
+{
+  return PyUnicode_FromFormat("<apsw.zeroblob object size %lld at %p>",
+                              self->blobsize,
+                              self);
+}
+
 static PyMethodDef ZeroBlobBind_methods[] = {
     {"length", (PyCFunction)ZeroBlobBind_len, METH_NOARGS,
      Zeroblob_length_DOC},
@@ -113,6 +114,7 @@ static PyTypeObject ZeroBlobBindType = {
     .tp_methods = ZeroBlobBind_methods,
     .tp_init = (initproc)ZeroBlobBind_init,
     .tp_new = ZeroBlobBind_new,
+    .tp_str = (reprfunc)ZeroBlobBind_tp_str,
 };
 
 /* BLOB TYPE */
@@ -134,9 +136,9 @@ static PyTypeObject APSWBlobType;
 
 /** .. class:: Blob
 
-  This object is created by :meth:`Connection.blobopen` and provides
+  This object is created by :meth:`Connection.blob_open` and provides
   access to a blob in the database.  It behaves like a Python file.
-  At the C level it wraps a `sqlite3_blob
+  It wraps a `sqlite3_blob
   <https://sqlite.org/c3ref/blob.html>`_.
 
   .. note::
@@ -163,10 +165,8 @@ static int
 APSWBlob_close_internal(APSWBlob *self, int force)
 {
   int setexc = 0;
-  PyObject *err_type, *err_value, *err_traceback;
 
-  if (force == 2)
-    PyErr_Fetch(&err_type, &err_value, &err_traceback);
+  PY_ERR_FETCH_IF(force == 2, exc_save);
 
   /* note that sqlite3_blob_close always works even if an error is
      returned */
@@ -202,7 +202,7 @@ APSWBlob_close_internal(APSWBlob *self, int force)
   Py_CLEAR(self->connection);
 
   if (force == 2)
-    PyErr_Restore(err_type, err_value, err_traceback);
+    PY_ERR_RESTORE(exc_save);
 
   return setexc;
 }
@@ -245,13 +245,13 @@ APSWBlob_length(APSWBlob *self)
   Reads amount of data requested, or till end of file, whichever is
   earlier. Attempting to read beyond the end of the blob returns an
   empty bytes in the same manner as end of file on normal file
-  objects.  Negative numbers read remaining data.
+  objects.  Negative numbers read all remaining data.
 
   -* sqlite3_blob_read
 */
 
 static PyObject *
-APSWBlob_read(APSWBlob *self, PyObject *args, PyObject *kwds)
+APSWBlob_read(APSWBlob *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int length = -1;
   int res;
@@ -267,10 +267,10 @@ APSWBlob_read(APSWBlob *self, PyObject *args, PyObject *kwds)
      why).  In any event we remain consistent with Python file
      objects */
   {
-    static char *kwlist[] = {"length", NULL};
     Blob_read_CHECK;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|i:" Blob_read_USAGE, kwlist, &length))
-      return NULL;
+    ARG_PROLOG(1, Blob_read_KWNAMES);
+    ARG_OPTIONAL ARG_int(length);
+    ARG_EPILOG(NULL, Blob_read_USAGE, );
   }
 
   if (
@@ -283,8 +283,7 @@ APSWBlob_read(APSWBlob *self, PyObject *args, PyObject *kwds)
     length = sqlite3_blob_bytes(self->pBlob) - self->curoffset;
 
   /* trying to read more than is in the blob? */
-  /* ::TODO:: use 64 arithmetic to avoid overflow */
-  if (self->curoffset + length > sqlite3_blob_bytes(self->pBlob))
+  if ((sqlite3_int64)self->curoffset + (sqlite3_int64)length > sqlite3_blob_bytes(self->pBlob))
     length = sqlite3_blob_bytes(self->pBlob) - self->curoffset;
 
   buffy = PyBytes_FromStringAndSize(NULL, length);
@@ -312,7 +311,7 @@ APSWBlob_read(APSWBlob *self, PyObject *args, PyObject *kwds)
   return buffy;
 }
 
-/** .. method:: readinto(buffer: bytearray |  array.array[Any] | memoryview, offset: int = 0, length: int = -1) -> None
+/** .. method:: read_into(buffer: bytearray |  array.array[Any] | memoryview, offset: int = 0, length: int = -1) -> None
 
   Reads from the blob into a buffer you have supplied.  This method is
   useful if you already have a buffer like object that data is being
@@ -335,7 +334,7 @@ APSWBlob_read(APSWBlob *self, PyObject *args, PyObject *kwds)
 */
 
 static PyObject *
-APSWBlob_readinto(APSWBlob *self, PyObject *args, PyObject *kwds)
+APSWBlob_read_into(APSWBlob *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int res = SQLITE_OK;
   long long offset = 0, length = -1;
@@ -349,10 +348,12 @@ APSWBlob_readinto(APSWBlob *self, PyObject *args, PyObject *kwds)
   CHECK_USE(NULL);
   CHECK_BLOB_CLOSED;
   {
-    static char *kwlist[] = {"buffer", "offset", "length", NULL};
-    Blob_readinto_CHECK;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|LL:" Blob_readinto_USAGE, kwlist, &buffer, &offset, &length))
-      return NULL;
+    Blob_read_into_CHECK;
+    ARG_PROLOG(3, Blob_read_into_KWNAMES);
+    ARG_MANDATORY ARG_pyobject(buffer);
+    ARG_OPTIONAL ARG_int64(offset);
+    ARG_OPTIONAL ARG_int64(length);
+    ARG_EPILOG(NULL, Blob_read_into_USAGE, );
   }
 
 #define ERREXIT(x)  \
@@ -363,7 +364,7 @@ APSWBlob_readinto(APSWBlob *self, PyObject *args, PyObject *kwds)
   } while (0)
 
   memset(&py3buffer, 0, sizeof(py3buffer));
-  aswb = PyObject_GetBuffer(buffer, &py3buffer, PyBUF_WRITABLE | PyBUF_SIMPLE);
+  aswb = PyObject_GetBufferContiguous(buffer, &py3buffer, PyBUF_WRITABLE | PyBUF_SIMPLE);
   if (aswb)
     return NULL;
 
@@ -385,7 +386,7 @@ APSWBlob_readinto(APSWBlob *self, PyObject *args, PyObject *kwds)
 
   MakeExistingException(); /* vfs errors could cause this */
 
-  if(PyErr_Occurred())
+  if (PyErr_Occurred())
     ERREXIT(NULL);
 
   if (res != SQLITE_OK)
@@ -416,17 +417,18 @@ errorexit:
 */
 
 static PyObject *
-APSWBlob_seek(APSWBlob *self, PyObject *args, PyObject *kwds)
+APSWBlob_seek(APSWBlob *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int offset, whence = 0;
   CHECK_USE(NULL);
   CHECK_BLOB_CLOSED;
 
   {
-    static char *kwlist[] = {"offset", "whence", NULL};
     Blob_seek_CHECK;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "i|i:" Blob_seek_USAGE, kwlist, &offset, &whence))
-      return NULL;
+    ARG_PROLOG(2, Blob_seek_KWNAMES);
+    ARG_MANDATORY ARG_int(offset);
+    ARG_OPTIONAL ARG_int(whence);
+    ARG_EPILOG(NULL, Blob_seek_USAGE, );
   }
   switch (whence)
   {
@@ -482,22 +484,29 @@ APSWBlob_tell(APSWBlob *self)
   -* sqlite3_blob_write
 */
 static PyObject *
-APSWBlob_write(APSWBlob *self, PyObject *args, PyObject *kwds)
+APSWBlob_write(APSWBlob *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int ok = 0, res = SQLITE_OK;
-  Py_buffer data;
+  Py_buffer data_buffer;
+  PyObject *data;
 
   CHECK_USE(NULL);
   CHECK_BLOB_CLOSED;
 
   {
-    static char *kwlist[] = {"data", NULL};
     Blob_write_CHECK;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "y*:" Blob_write_USAGE, kwlist, &data))
-      return NULL;
+    ARG_PROLOG(1, Blob_write_KWNAMES);
+    ARG_MANDATORY ARG_py_buffer(data);
+    ARG_EPILOG(NULL, Blob_write_USAGE, );
   }
 
-  Py_ssize_t calc_end = data.len + self->curoffset;
+  if (0 != PyObject_GetBufferContiguous(data, &data_buffer, PyBUF_SIMPLE))
+  {
+    assert(PyErr_Occurred());
+    return NULL;
+  }
+
+  Py_ssize_t calc_end = data_buffer.len + self->curoffset;
 
   APSW_FAULT_INJECT(BlobWriteTooBig, , calc_end = (Py_ssize_t)0x7FFFFFFF * (Py_ssize_t)0x1000);
 
@@ -513,7 +522,7 @@ APSWBlob_write(APSWBlob *self, PyObject *args, PyObject *kwds)
     goto finally;
   }
 
-  PYSQLITE_BLOB_CALL(res = sqlite3_blob_write(self->pBlob, data.buf, data.len, self->curoffset));
+  PYSQLITE_BLOB_CALL(res = sqlite3_blob_write(self->pBlob, data_buffer.buf, data_buffer.len, self->curoffset));
   assert(!PyErr_Occurred());
 
   if (res != SQLITE_OK)
@@ -521,12 +530,12 @@ APSWBlob_write(APSWBlob *self, PyObject *args, PyObject *kwds)
     SET_EXC(res, self->connection->db);
     goto finally;
   }
-  self->curoffset += data.len;
+  self->curoffset += data_buffer.len;
   assert(self->curoffset <= sqlite3_blob_bytes(self->pBlob));
   ok = 1;
 
 finally:
-  PyBuffer_Release(&data);
+  PyBuffer_Release(&data_buffer);
   if (ok)
     Py_RETURN_NONE;
   else
@@ -556,7 +565,7 @@ finally:
 */
 
 static PyObject *
-APSWBlob_close(APSWBlob *self, PyObject *args, PyObject *kwds)
+APSWBlob_close(APSWBlob *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int setexc;
   int force = 0;
@@ -564,11 +573,10 @@ APSWBlob_close(APSWBlob *self, PyObject *args, PyObject *kwds)
   CHECK_USE(NULL);
 
   {
-    static char *kwlist[] = {"force", NULL};
     Blob_close_CHECK;
-    argcheck_bool_param force_param = {&force, Blob_close_force_MSG};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O&:" Blob_close_USAGE, kwlist, argcheck_bool, &force_param))
-      return NULL;
+    ARG_PROLOG(1, Blob_close_KWNAMES);
+    ARG_OPTIONAL ARG_bool(force);
+    ARG_EPILOG(NULL, Blob_close_USAGE, );
   }
   setexc = APSWBlob_close_internal(self, !!force);
 
@@ -581,14 +589,14 @@ APSWBlob_close(APSWBlob *self, PyObject *args, PyObject *kwds)
 /** .. method:: __enter__() -> Blob
 
   You can use a blob as a `context manager
-  <https://docs.python.org/reference/datamodel.html#with-statement-context-managers>`_
+  <https://docs.python.org/3/reference/datamodel.html#with-statement-context-managers>`_
   as defined in :pep:`0343`.  When you use *with* statement,
   the blob is always :meth:`closed <Blob.close>` on exit from the block, even if an
   exception occurred in the block.
 
   For example::
 
-    with connection.blobopen() as blob:
+    with connection.blob_open() as blob:
         blob.write("...")
         res=blob.read(1024)
 
@@ -633,7 +641,7 @@ APSWBlob_exit(APSWBlob *self, PyObject *Py_UNUSED(args))
 */
 
 static PyObject *
-APSWBlob_reopen(APSWBlob *self, PyObject *args, PyObject *kwds)
+APSWBlob_reopen(APSWBlob *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int res;
   long long rowid;
@@ -642,10 +650,10 @@ APSWBlob_reopen(APSWBlob *self, PyObject *args, PyObject *kwds)
   CHECK_BLOB_CLOSED;
 
   {
-    static char *kwlist[] = {"rowid", NULL};
     Blob_reopen_CHECK;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "L:" Blob_reopen_USAGE, kwlist, &rowid))
-      return NULL;
+    ARG_PROLOG(1, Blob_reopen_KWNAMES);
+    ARG_MANDATORY ARG_int64(rowid);
+    ARG_EPILOG(NULL, Blob_reopen_USAGE, );
   }
   /* no matter what happens we always reset current offset */
   self->curoffset = 0;
@@ -665,27 +673,39 @@ APSWBlob_reopen(APSWBlob *self, PyObject *args, PyObject *kwds)
   Py_RETURN_NONE;
 }
 
+static PyObject *
+APSWBlob_tp_str(APSWBlob *self)
+{
+  return PyUnicode_FromFormat("<apsw.Blob object from %S at %p>",
+                              self->connection ? (PyObject *)self->connection : apst.closed,
+                              self);
+}
+
 static PyMethodDef APSWBlob_methods[] = {
     {"length", (PyCFunction)APSWBlob_length, METH_NOARGS,
      Blob_length_DOC},
-    {"read", (PyCFunction)APSWBlob_read, METH_VARARGS | METH_KEYWORDS,
+    {"read", (PyCFunction)APSWBlob_read, METH_FASTCALL | METH_KEYWORDS,
      Blob_read_DOC},
-    {"readinto", (PyCFunction)APSWBlob_readinto, METH_VARARGS | METH_KEYWORDS,
-     Blob_readinto_DOC},
-    {"seek", (PyCFunction)APSWBlob_seek, METH_VARARGS | METH_KEYWORDS,
+    {"read_into", (PyCFunction)APSWBlob_read_into, METH_FASTCALL | METH_KEYWORDS,
+     Blob_read_into_DOC},
+    {"seek", (PyCFunction)APSWBlob_seek, METH_FASTCALL | METH_KEYWORDS,
      Blob_seek_DOC},
     {"tell", (PyCFunction)APSWBlob_tell, METH_NOARGS,
      Blob_tell_DOC},
-    {"write", (PyCFunction)APSWBlob_write, METH_VARARGS | METH_KEYWORDS,
+    {"write", (PyCFunction)APSWBlob_write, METH_FASTCALL | METH_KEYWORDS,
      Blob_write_DOC},
-    {"reopen", (PyCFunction)APSWBlob_reopen, METH_VARARGS | METH_KEYWORDS,
+    {"reopen", (PyCFunction)APSWBlob_reopen, METH_FASTCALL | METH_KEYWORDS,
      Blob_reopen_DOC},
-    {"close", (PyCFunction)APSWBlob_close, METH_VARARGS | METH_KEYWORDS,
+    {"close", (PyCFunction)APSWBlob_close, METH_FASTCALL | METH_KEYWORDS,
      Blob_close_DOC},
     {"__enter__", (PyCFunction)APSWBlob_enter, METH_NOARGS,
      Blob_enter_DOC},
     {"__exit__", (PyCFunction)APSWBlob_exit, METH_VARARGS,
      Blob_exit_DOC},
+#ifndef APSW_OMIT_OLD_NAMES
+    {Blob_read_into_OLDNAME, (PyCFunction)APSWBlob_read_into, METH_FASTCALL | METH_KEYWORDS,
+     Blob_read_into_OLDDOC},
+#endif
     {0, 0, 0, 0} /* Sentinel */
 };
 
@@ -698,4 +718,5 @@ static PyTypeObject APSWBlobType = {
     .tp_doc = Blob_class_DOC,
     .tp_weaklistoffset = offsetof(APSWBlob, weakreflist),
     .tp_methods = APSWBlob_methods,
+    .tp_str = (reprfunc)APSWBlob_tp_str,
 };
