@@ -19,87 +19,17 @@ obtain a cursor behind the scenes.
 If you need a cursor you should call :meth:`~Connection.cursor` on your
 database::
 
-  db=apsw.Connection("databasefilename")
-  cursor=db.cursor()
+  db = apsw.Connection("databasefilename")
+  cursor = db.cursor()
 
-A cursor executes SQL::
+The :ref:`example <example_executing_sql>` shows how to execute SQL and
+how to provide values used in queries (bindings).
 
-  cursor.execute("create table example(title, isbn)")
-
-You can also read data back.  The row is returned as a tuple of the
-column values::
-
-  for row in cursor.execute("select * from example"):
-     print(row)
-
-There are two ways of supplying data to a query.  The **really bad** way is to compose a string::
-
-  sql="insert into example values('%s', %d)" % ("string", 8390823904)
-  cursor.execute(sql)
-
-If there were any single quotes in string then you would have invalid
-syntax.  Additionally this is how `SQL injection attacks
-<https://en.wikipedia.org/wiki/SQL_injection>`_ happen. Instead you should use bindings::
-
-  sql="insert into example values(?, ?)"
-  cursor.execute(sql, ("string", 8390823904))
-
-  # You can also use dictionaries (with colon, $, or @ before names)
-  sql="insert into example values(:title, :isbn)"
-  cursor.execute(sql, {"title": "string", "isbn": 8390823904})
-
-  # You can use local variables as the dictionary
-  title="..."
-  isbn="...."
-  cursor.execute(sql, locals())
-
-Cursors are cheap.  Use as many as you need.  It is safe to use them
-across threads, such as calling :meth:`~Cursor.execute` in one thread,
-passing the cursor to another thread that then calls
-`next <https://docs.python.org/3/library/functions.html?highlight=next#next>`__.  The only thing you can't do is call methods at
-exactly the same time on the same cursor in two different threads - eg
-trying to call :meth:`~Cursor.execute` in both at the same time, or
-:meth:`~Cursor.execute` in one and `next <https://docs.python.org/3/library/functions.html?highlight=next#next>`__ in another.
-(If you do attempt this, it will be detected and
-:exc:`ThreadingViolationError` will be raised.)
-
-Behind the scenes a :class:`Cursor` maps to a `SQLite statement
-<https://sqlite.org/c3ref/stmt.html>`_.  APSW maintains a
-:ref:`cache <statementcache>` so that the mapping is very fast, and the
+Cursors are cheap.  Use as many as you need. Behind the scenes a
+:class:`Cursor` maps to a `SQLite statement <https://sqlite.org/c3ref/stmt.html>`_.
+APSW maintains a :ref:`cache <statementcache>` so that the mapping is very fast, and the
 SQLite objects are reused when possible.
 
-A unique feature of APSW is that your query can be multiple semi-colon
-separated statements.  For example::
-
-  cursor.execute("select ... ; insert into ... ; update ... ; select ...")
-
-.. note::
-
-  SQLite fetches data as it is needed.  If table *example* had 10
-  million rows it would only get the next row as requested.  This
-  code would not work as expected::
-
-    for row in cursor.execute("select * from example"):
-       cursor.execute("insert .....")
-
-  The nested :meth:`~Cursor.execute` would start a new query
-  abandoning any remaining results from the ``SELECT`` cursor.  There are two
-  ways to work around this.  Use a different cursor::
-
-    for row in cursor1.execute("select * from example"):
-       cursor2.execute("insert ...")
-
-  You can also get all the rows immediately by filling in a list::
-
-    rows=list( cursor.execute("select * from example") )
-    for row in rows:
-       cursor.execute("insert ...")
-
-  This last approach is recommended since you don't have to worry
-  about the database changing while doing the ``select``.  You should
-  also understand transactions and where to put the transaction
-  boundaries.
-
 .. note::
 
   Cursors on the same :ref:`Connection <connections>` are not isolated
@@ -113,14 +43,14 @@ separated statements.  For example::
 
   * `SQLite transactions <https://sqlite.org/lang_transaction.html>`_
   * `Atomic commit <https://sqlite.org/atomiccommit.html>`_
-  * `Example of changing the database while running a query problem <https://www.mail-archive.com/sqlite-users@sqlite.org/msg42660.html>`_
   * :ref:`Benchmarking`
 
 */
 
-/** .. class:: Cursor
+/* secret backdoor to allow bindings all as null */
+static PyObject *apsw_cursor_null_bindings;
 
-  You obtain cursors by calling :meth:`Connection.cursor`.
+/** .. class:: Cursor
 */
 
 /* CURSOR TYPE */
@@ -158,6 +88,8 @@ struct APSWCursor
   PyObject *weakreflist;
 
   PyObject *description_cache[3];
+
+  int init_was_called;
 };
 
 typedef struct APSWCursor APSWCursor;
@@ -178,23 +110,23 @@ static int
 resetcursor(APSWCursor *self, int force)
 {
   int res = SQLITE_OK;
-  PyObject *etype, *eval, *etb;
   int hasmore = statementcache_hasmore(self->statement);
 
   Py_CLEAR(self->description_cache[0]);
   Py_CLEAR(self->description_cache[1]);
   Py_CLEAR(self->description_cache[2]);
 
-  if (force)
-    PyErr_Fetch(&etype, &eval, &etb);
+  PY_ERR_FETCH_IF(force, exc_save);
 
   if (self->statement)
   {
     INUSE_CALL(res = statementcache_finalize(self->connection->stmtcache, self->statement));
+    if (res == SQLITE_OK && PyErr_Occurred())
+      res = SQLITE_ERROR;
     if (res)
     {
-      if (force)
-        PyErr_Clear();
+      if (force && PyErr_Occurred())
+        apsw_write_unraisable(NULL);
       else
         SET_EXC(res, self->connection->db);
     }
@@ -241,7 +173,7 @@ resetcursor(APSWCursor *self, int force)
   }
 
   if (force)
-    PyErr_Restore(etype, eval, etb);
+    PY_ERR_RESTORE(exc_save);
 
   return res;
 }
@@ -249,16 +181,14 @@ resetcursor(APSWCursor *self, int force)
 static int
 APSWCursor_close_internal(APSWCursor *self, int force)
 {
-  PyObject *err_type, *err_value, *err_traceback;
   int res;
 
-  if (force == 2)
-    PyErr_Fetch(&err_type, &err_value, &err_traceback);
+  PY_ERR_FETCH_IF(force == 2, exc_save);
 
   res = resetcursor(self, force);
 
   if (force == 2)
-    PyErr_Restore(err_type, err_value, err_traceback);
+    PY_ERR_RESTORE(exc_save);
   else
   {
     if (res)
@@ -296,8 +226,7 @@ APSWCursor_dealloc(APSWCursor *self)
 {
   /* dealloc is not allowed to return an exception or
      clear the current exception */
-  PyObject *one, *two, *three;
-  PyErr_Fetch(&one, &two, &three);
+  PY_ERR_FETCH(exc_save);
 
   PyObject_GC_UnTrack(self);
   APSW_CLEAR_WEAKREFS;
@@ -307,7 +236,7 @@ APSWCursor_dealloc(APSWCursor *self)
   if (PyErr_Occurred())
     apsw_write_unraisable(NULL);
 
-  PyErr_Restore(one, two, three);
+  PY_ERR_RESTORE(exc_save);
   Py_TpFree((PyObject *)self);
 }
 
@@ -333,6 +262,7 @@ APSWCursor_new(PyTypeObject *type, PyObject *Py_UNUSED(args), PyObject *Py_UNUSE
     self->description_cache[0] = 0;
     self->description_cache[1] = 0;
     self->description_cache[2] = 0;
+    self->init_was_called = 0;
   }
 
   return (PyObject *)self;
@@ -345,15 +275,17 @@ APSWCursor_new(PyTypeObject *type, PyObject *Py_UNUSED(args), PyObject *Py_UNUSE
 */
 
 static int
-APSWCursor_init(APSWCursor *self, PyObject *args, PyObject *kwds)
+APSWCursor_init(APSWCursor *self, PyObject *args, PyObject *kwargs)
 {
   Connection *connection = NULL;
 
   {
-    static char *kwlist[] = {"connection", NULL};
     Cursor_init_CHECK;
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!:" Cursor_init_USAGE, kwlist, &ConnectionType, &connection))
-      return -1;
+    PREVENT_INIT_MULTIPLE_CALLS;
+    ARG_CONVERT_VARARGS_TO_FASTCALL;
+    ARG_PROLOG(1, Cursor_init_KWNAMES);
+    ARG_MANDATORY ARG_Connection(connection);
+    ARG_EPILOG(-1, Cursor_init_USAGE, Py_XDECREF(fast_kwnames));
   }
 
   self->connection = (Connection *)Py_NewRef((PyObject *)connection);
@@ -376,7 +308,7 @@ static const char *description_formats[] = {
     "(sssss)"};
 
 static PyObject *
-APSWCursor_internal_getdescription(APSWCursor *self, int fmtnum)
+APSWCursor_internal_get_description(APSWCursor *self, int fmtnum)
 {
   int ncols, i;
   PyObject *result = NULL;
@@ -453,63 +385,29 @@ error:
   return NULL;
 }
 
-/** .. method:: getdescription() -> tuple[tuple[str, str], ...]
+/** .. method:: get_description() -> tuple[tuple[str, str], ...]
 
    If you are trying to get information about a table or view,
    then `pragma table_info <https://sqlite.org/pragma.html#pragma_table_info>`__
-   is better.
+   is better.  If you want to know up front what columns and other
+   details a query does then :func:`apsw.ext.query_info` is useful.
 
    Returns a tuple describing each column in the result row.  The
-   return is identical for every row of the results.  You can only
-   call this method once you have started executing a statement and
-   before you have finished::
-
-      # This will error
-      cursor.getdescription()
-
-      for row in cursor.execute("select ....."):
-         # this works
-         print (cursor.getdescription())
-         print (row)
+   return is identical for every row of the results.
 
    The information about each column is a tuple of ``(column_name,
    declared_column_type)``.  The type is what was declared in the
    ``CREATE TABLE`` statement - the value returned in the row will be
-   whatever type you put in for that row and column.  (This is known
-   as `manifest typing <https://sqlite.org/different.html#typing>`_
-   which is also the way that Python works.  The variable ``a`` could
-   contain an integer, and then you could put a string in it.  Other
-   static languages such as C or other SQL databases only let you put
-   one type in - eg ``a`` could only contain an integer or a string,
-   but never both.)
+   whatever type you put in for that row and column.
 
-   Example::
-
-      cursor.execute("create table books(title string, isbn number, wibbly wobbly zebra)")
-      cursor.execute("insert into books values(?,?,?)", (97, "fjfjfj", 3.7))
-      cursor.execute("insert into books values(?,?,?)", ("fjfjfj", 3.7, 97))
-
-      for row in cursor.execute("select * from books"):
-         print (cursor.getdescription())
-         print (row)
-
-   Output::
-
-     # row 0 - description
-     (('title', 'string'), ('isbn', 'number'), ('wibbly', 'wobbly zebra'))
-     # row 0 - values
-     (97, 'fjfjfj', 3.7)
-     # row 1 - description
-     (('title', 'string'), ('isbn', 'number'), ('wibbly', 'wobbly zebra'))
-     # row 1 - values
-     ('fjfjfj', 3.7, 97)
+   See the :ref:`query_info example <example_query_details>`.
 
    -* sqlite3_column_name sqlite3_column_decltype
 
 */
-static PyObject *APSWCursor_getdescription(APSWCursor *self)
+static PyObject *APSWCursor_get_description(APSWCursor *self)
 {
-  return APSWCursor_internal_getdescription(self, 0);
+  return APSWCursor_internal_get_description(self, 0);
 }
 
 /** .. attribute:: description
@@ -517,13 +415,13 @@ static PyObject *APSWCursor_getdescription(APSWCursor *self)
 
     Based on the `DB-API cursor property
     <https://www.python.org/dev/peps/pep-0249/>`__, this returns the
-    same as :meth:`getdescription` but with 5 Nones appended.  See
-    also :issue:`131`.
+    same as :meth:`get_description` but with 5 Nones appended because
+    SQLite does not have the information.
 */
 
 static PyObject *APSWCursor_getdescription_dbapi(APSWCursor *self)
 {
-  return APSWCursor_internal_getdescription(self, 1);
+  return APSWCursor_internal_get_description(self, 1);
 }
 
 /** .. attribute:: description_full
@@ -540,9 +438,9 @@ name, table name, and origin name.
 
 */
 #ifdef SQLITE_ENABLE_COLUMN_METADATA
-static PyObject *APSWCursor_getdescription_full(APSWCursor *self)
+static PyObject *APSWCursor_get_description_full(APSWCursor *self)
 {
-  return APSWCursor_internal_getdescription(self, 2);
+  return APSWCursor_internal_get_description(self, 2);
 }
 #endif
 
@@ -617,7 +515,7 @@ APSWCursor_dobinding(APSWCursor *self, int arg, PyObject *obj)
     int asrb;
     Py_buffer py3buffer;
 
-    asrb = PyObject_GetBuffer(obj, &py3buffer, PyBUF_SIMPLE);
+    asrb = PyObject_GetBufferContiguous(obj, &py3buffer, PyBUF_SIMPLE);
     if (asrb != 0)
       return -1;
 
@@ -653,6 +551,10 @@ APSWCursor_dobindings(APSWCursor *self)
   assert(!PyErr_Occurred());
   assert(self->bindingsoffset >= 0);
 
+  /* skip for null bindings */
+  if (Py_Is(self->bindings, apsw_cursor_null_bindings))
+    return 0;
+
   nargs = sqlite3_bind_parameter_count(self->statement->vdbestatement);
   if (nargs == 0 && !self->bindings)
     return 0; /* common case, no bindings needed or supplied */
@@ -670,7 +572,7 @@ APSWCursor_dobindings(APSWCursor *self)
     {
       const char *key;
 
-      PYSQLITE_CUR_CALL(key = sqlite3_bind_parameter_name(self->statement->vdbestatement, arg));
+      key = sqlite3_bind_parameter_name(self->statement->vdbestatement, arg);
 
       if (!key)
       {
@@ -750,7 +652,7 @@ APSWCursor_dobindings(APSWCursor *self)
 }
 
 static int
-APSWCursor_doexectrace(APSWCursor *self, Py_ssize_t savedbindingsoffset)
+APSWCursor_do_exec_trace(APSWCursor *self, Py_ssize_t savedbindingsoffset)
 {
   PyObject *retval = NULL;
   PyObject *sqlcmd = NULL;
@@ -775,6 +677,10 @@ APSWCursor_doexectrace(APSWCursor *self, Py_ssize_t savedbindingsoffset)
     {
       bindings = Py_NewRef(self->bindings);
     }
+    else if (Py_Is(self->bindings, apsw_cursor_null_bindings))
+    {
+      bindings = Py_NewRef(Py_None);
+    }
     else
     {
       bindings = PySequence_GetSlice(self->bindings, savedbindingsoffset, self->bindingsoffset);
@@ -791,7 +697,10 @@ APSWCursor_doexectrace(APSWCursor *self, Py_ssize_t savedbindingsoffset)
     bindings = Py_NewRef(Py_None);
   }
 
-  retval = PyObject_CallFunction(exectrace, "ONN", self, sqlcmd, bindings);
+  PyObject *vargs[] = {NULL, (PyObject *)self, sqlcmd, bindings};
+  retval = PyObject_Vectorcall(exectrace, vargs + 1, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
+  Py_DECREF(sqlcmd);
+  Py_DECREF(bindings);
 
   if (!retval)
   {
@@ -815,13 +724,14 @@ APSWCursor_doexectrace(APSWCursor *self, Py_ssize_t savedbindingsoffset)
 }
 
 static PyObject *
-APSWCursor_dorowtrace(APSWCursor *self, PyObject *retval)
+APSWCursor_do_row_trace(APSWCursor *self, PyObject *retval)
 {
   PyObject *rowtrace = ROWTRACE;
 
   assert(rowtrace);
 
-  return PyObject_CallFunction(rowtrace, "OO", self, retval);
+  PyObject *vargs[] = {NULL, (PyObject *)self, retval};
+  return PyObject_Vectorcall(rowtrace, vargs + 1, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
 }
 
 /* Returns a borrowed reference to self if all is ok, else NULL on error */
@@ -908,7 +818,7 @@ APSWCursor_step(APSWCursor *self)
         self->bindings = next;
       else
       {
-        self->bindings = PySequence_Fast(next, "You must supply a dict or a sequence");
+        self->bindings = PySequence_Fast(next, "You must supply a dict or a sequence for bindings");
         /* we no longer need next irrespective of what happens in line above */
         Py_DECREF(next);
         if (!self->bindings)
@@ -956,7 +866,7 @@ APSWCursor_step(APSWCursor *self)
 
     if (EXECTRACE)
     {
-      if (APSWCursor_doexectrace(self, savedbindingsoffset))
+      if (APSWCursor_do_exec_trace(self, savedbindingsoffset))
       {
         assert(self->status == C_DONE);
         assert(PyErr_Occurred());
@@ -989,37 +899,6 @@ APSWCursor_step(APSWCursor *self)
     :param explain: If 0 or greater then the statement is passed to `sqlite3_stmt_explain <https://sqlite.org/c3ref/stmt_explain.html>`__
        where you can force it to not be an explain, or force explain or explain query plan.
 
-    If you use numbered bindings in the query then supply a sequence.
-    Any sequence will work including lists and iterators.  For
-    example::
-
-      cursor.execute("insert into books values(?,?)", ("title", "number"))
-
-    .. note::
-
-      A common gotcha is wanting to insert a single string but not
-      putting it in a tuple::
-
-        cursor.execute("insert into books values(?)", "a title")
-
-      The string is a sequence of 8 characters and so it will look
-      like you are supplying 8 bindings when only one is needed.  Use
-      a one item tuple with a trailing comma like this::
-
-        cursor.execute("insert into books values(?)", ("a title",) )
-
-    If you used names in the statement then supply a dictionary as the
-    binding.  It is ok to be missing entries from the dictionary -
-    None/null will be used.  For example::
-
-       cursor.execute("insert into books values(:title, :isbn, :rating)",
-            {"title": "book title", "isbn": 908908908})
-
-    The return is the cursor object itself which is also an iterator.  This allows you to write::
-
-       for row in cursor.execute("select * from books"):
-          print(row)
-
     :raises TypeError: The bindings supplied were neither a dict nor a sequence
     :raises BindingsError: You supplied too many or too few bindings for the statements
     :raises IncompleteExecutionError: There are remaining unexecuted queries from your last execute
@@ -1028,11 +907,12 @@ APSWCursor_step(APSWCursor *self)
 
     .. seealso::
 
+       * :ref:`Example <example_executing_sql>` showing how to use bindings
        * :ref:`executionmodel`
 
 */
 static PyObject *
-APSWCursor_execute(APSWCursor *self, PyObject *args, PyObject *kwds)
+APSWCursor_execute(APSWCursor *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int res;
   int savedbindingsoffset = -1;
@@ -1055,12 +935,14 @@ APSWCursor_execute(APSWCursor *self, PyObject *args, PyObject *kwds)
 
   assert(!self->bindings);
   {
-    static char *kwlist[] = {"statements", "bindings", "can_cache", "prepare_flags", "explain", NULL};
     Cursor_execute_CHECK;
-    argcheck_Optional_Bindings_param bindings_param = {&bindings, Cursor_execute_bindings_MSG};
-    argcheck_bool_param can_cache_param = {&can_cache, Cursor_execute_can_cache_MSG};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!|O&$O&ii:" Cursor_execute_USAGE, kwlist, &PyUnicode_Type, &statements, argcheck_Optional_Bindings, &bindings_param, argcheck_bool, &can_cache_param, &prepare_flags, &explain))
-      return NULL;
+    ARG_PROLOG(2, Cursor_execute_KWNAMES);
+    ARG_MANDATORY ARG_PyUnicode(statements);
+    ARG_OPTIONAL ARG_optional_Bindings(bindings);
+    ARG_OPTIONAL ARG_bool(can_cache);
+    ARG_OPTIONAL ARG_int(prepare_flags);
+    ARG_OPTIONAL ARG_int(explain);
+    ARG_EPILOG(NULL, Cursor_execute_USAGE, );
   }
   self->bindings = bindings;
 
@@ -1070,11 +952,11 @@ APSWCursor_execute(APSWCursor *self, PyObject *args, PyObject *kwds)
 
   if (self->bindings)
   {
-    if (APSWCursor_is_dict_binding(self->bindings))
+    if (APSWCursor_is_dict_binding(self->bindings) || Py_Is(self->bindings, apsw_cursor_null_bindings))
       Py_INCREF(self->bindings);
     else
     {
-      self->bindings = PySequence_Fast(self->bindings, "You must supply a dict or a sequence");
+      self->bindings = PySequence_Fast(self->bindings, "You must supply a dict or a sequence for execute");
       if (!self->bindings)
         return NULL;
     }
@@ -1103,7 +985,7 @@ APSWCursor_execute(APSWCursor *self, PyObject *args, PyObject *kwds)
 
   if (EXECTRACE)
   {
-    if (APSWCursor_doexectrace(self, savedbindingsoffset))
+    if (APSWCursor_do_exec_trace(self, savedbindingsoffset))
     {
       assert(PyErr_Occurred());
       return NULL;
@@ -1121,7 +1003,7 @@ APSWCursor_execute(APSWCursor *self, PyObject *args, PyObject *kwds)
   return Py_NewRef(retval);
 }
 
-/** .. method:: executemany(statements: str, sequenceofbindings: Sequence[Bindings], *, can_cache: bool = True, prepare_flags: int = 0, explain: int = -1) -> Cursor
+/** .. method:: executemany(statements: str, sequenceofbindings: Iterable[Bindings], *, can_cache: bool = True, prepare_flags: int = 0, explain: int = -1) -> Cursor
 
   This method is for when you want to execute the same statements over
   a sequence of bindings.  Conceptually it does this::
@@ -1129,22 +1011,14 @@ APSWCursor_execute(APSWCursor *self, PyObject *args, PyObject *kwds)
     for binding in sequenceofbindings:
         cursor.execute(statements, binding)
 
-  Example::
-
-    rows=(  (1, 7),
-            (2, 23),
-            (4, 92),
-            (12, 12) )
-
-    cursor.executemany("insert into nums values(?,?)", rows)
-
   The return is the cursor itself which acts as an iterator.  Your
   statements can return data.  See :meth:`~Cursor.execute` for more
-  information.
+  information, and the :ref:`example <example_executemany>`.
+
 */
 
 static PyObject *
-APSWCursor_executemany(APSWCursor *self, PyObject *args, PyObject *kwds)
+APSWCursor_executemany(APSWCursor *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int res;
   PyObject *retval = NULL;
@@ -1171,11 +1045,14 @@ APSWCursor_executemany(APSWCursor *self, PyObject *args, PyObject *kwds)
   assert(!self->emoriginalquery);
   assert(self->status == C_DONE);
   {
-    static char *kwlist[] = {"statements", "sequenceofbindings", "can_cache", "prepare_flags", "explain", NULL};
     Cursor_executemany_CHECK;
-    argcheck_bool_param can_cache_param = {&can_cache, Cursor_executemany_can_cache_MSG};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!O|$O&ii:" Cursor_executemany_USAGE, kwlist, &PyUnicode_Type, &statements, &sequenceofbindings, argcheck_bool, &can_cache_param, &prepare_flags, &explain))
-      return NULL;
+    ARG_PROLOG(2, Cursor_executemany_KWNAMES);
+    ARG_MANDATORY ARG_PyUnicode(statements);
+    ARG_MANDATORY ARG_pyobject(sequenceofbindings);
+    ARG_OPTIONAL ARG_bool(can_cache);
+    ARG_OPTIONAL ARG_int(prepare_flags);
+    ARG_OPTIONAL ARG_int(explain);
+    ARG_EPILOG(NULL, Cursor_executemany_USAGE, );
   }
   self->emiter = PyObject_GetIter(sequenceofbindings);
   if (!self->emiter)
@@ -1197,7 +1074,7 @@ APSWCursor_executemany(APSWCursor *self, PyObject *args, PyObject *kwds)
     self->bindings = next;
   else
   {
-    self->bindings = PySequence_Fast(next, "You must supply a dict or a sequence");
+    self->bindings = PySequence_Fast(next, "You must supply a dict or a sequence for executemany");
     Py_DECREF(next); /* _Fast makes new reference */
     if (!self->bindings)
       return NULL;
@@ -1233,7 +1110,7 @@ APSWCursor_executemany(APSWCursor *self, PyObject *args, PyObject *kwds)
 
   if (EXECTRACE)
   {
-    if (APSWCursor_doexectrace(self, savedbindingsoffset))
+    if (APSWCursor_do_exec_trace(self, savedbindingsoffset))
     {
       assert(PyErr_Occurred());
       return NULL;
@@ -1253,16 +1130,14 @@ APSWCursor_executemany(APSWCursor *self, PyObject *args, PyObject *kwds)
 
 /** .. method:: close(force: bool = False) -> None
 
-  It is very unlikely you will need to call this method.  It exists
-  because older versions of SQLite required all Connection/Cursor
-  activity to be confined to the same thread.  That is no longer the
-  case.  Cursors are automatically garbage collected and when there
+  It is very unlikely you will need to call this method.
+  Cursors are automatically garbage collected and when there
   are none left will allow the connection to be garbage collected if
   it has no other references.
 
   A cursor is open if there are remaining statements to execute (if
   your query included multiple statements), or if you called
-  :meth:`~Cursor.executemany` and not all of the *sequenceofbindings*
+  :meth:`~Cursor.executemany` and not all of the sequence of bindings
   have been used yet.
 
   :param force: If False then you will get exceptions if there is
@@ -1274,7 +1149,7 @@ APSWCursor_executemany(APSWCursor *self, PyObject *args, PyObject *kwds)
 */
 
 static PyObject *
-APSWCursor_close(APSWCursor *self, PyObject *args, PyObject *kwds)
+APSWCursor_close(APSWCursor *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   int force = 0;
 
@@ -1283,11 +1158,10 @@ APSWCursor_close(APSWCursor *self, PyObject *args, PyObject *kwds)
     Py_RETURN_NONE;
 
   {
-    static char *kwlist[] = {"force", NULL};
     Cursor_close_CHECK;
-    argcheck_bool_param force_param = {&force, Cursor_close_force_MSG};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O&:" Cursor_close_USAGE, kwlist, argcheck_bool, &force_param))
-      return NULL;
+    ARG_PROLOG(1, Cursor_close_KWNAMES);
+    ARG_OPTIONAL ARG_bool(force);
+    ARG_EPILOG(NULL, Cursor_close_USAGE, );
   }
   APSWCursor_close_internal(self, !!force);
 
@@ -1341,7 +1215,7 @@ again:
   }
   if (ROWTRACE)
   {
-    PyObject *r2 = APSWCursor_dorowtrace(self, retval);
+    PyObject *r2 = APSWCursor_do_row_trace(self, retval);
     Py_DECREF(retval);
     if (!r2)
       return NULL;
@@ -1372,23 +1246,22 @@ APSWCursor_iter(APSWCursor *self)
   return Py_NewRef((PyObject *)self);
 }
 
-/** .. method:: setexectrace(callable: Optional[ExecTracer]) -> None
+/** .. method:: set_exec_trace(callable: Optional[ExecTracer]) -> None
 
-  Sets the :attr:`execution tracer <Cursor.exectrace>`
+  Sets the :attr:`execution tracer <Cursor.exec_trace>`
 */
 static PyObject *
-APSWCursor_setexectrace(APSWCursor *self, PyObject *args, PyObject *kwds)
+APSWCursor_set_exec_trace(APSWCursor *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   PyObject *callable = NULL;
   CHECK_USE(NULL);
   CHECK_CURSOR_CLOSED(NULL);
 
   {
-    static char *kwlist[] = {"callable", NULL};
-    Cursor_setexectrace_CHECK;
-    argcheck_Optional_Callable_param callable_param = {&callable, Cursor_setexectrace_callable_MSG};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&:" Cursor_setexectrace_USAGE, kwlist, argcheck_Optional_Callable, &callable_param))
-      return NULL;
+    Cursor_set_exec_trace_CHECK;
+    ARG_PROLOG(1, Cursor_set_exec_trace_KWNAMES);
+    ARG_MANDATORY ARG_optional_Callable(callable);
+    ARG_EPILOG(NULL, Cursor_set_exec_trace_USAGE, );
   }
 
   Py_XINCREF(callable);
@@ -1398,24 +1271,23 @@ APSWCursor_setexectrace(APSWCursor *self, PyObject *args, PyObject *kwds)
   Py_RETURN_NONE;
 }
 
-/** .. method:: setrowtrace(callable: Optional[RowTracer]) -> None
+/** .. method:: set_row_trace(callable: Optional[RowTracer]) -> None
 
-  Sets the :attr:`row tracer <Cursor.rowtrace>`
+  Sets the :attr:`row tracer <Cursor.row_trace>`
 */
 
 static PyObject *
-APSWCursor_setrowtrace(APSWCursor *self, PyObject *args, PyObject *kwds)
+APSWCursor_set_row_trace(APSWCursor *self, PyObject *const *fast_args, Py_ssize_t fast_nargs, PyObject *fast_kwnames)
 {
   PyObject *callable = NULL;
   CHECK_USE(NULL);
   CHECK_CURSOR_CLOSED(NULL);
 
   {
-    static char *kwlist[] = {"callable", NULL};
-    Cursor_setrowtrace_CHECK;
-    argcheck_Optional_Callable_param callable_param = {&callable, Cursor_setrowtrace_callable_MSG};
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&:" Cursor_setrowtrace_USAGE, kwlist, argcheck_Optional_Callable, &callable_param))
-      return NULL;
+    Cursor_set_row_trace_CHECK;
+    ARG_PROLOG(1, Cursor_set_row_trace_KWNAMES);
+    ARG_MANDATORY ARG_optional_Callable(callable);
+    ARG_EPILOG(NULL, Cursor_set_row_trace_USAGE, );
   }
 
   Py_XINCREF(callable);
@@ -1425,17 +1297,17 @@ APSWCursor_setrowtrace(APSWCursor *self, PyObject *args, PyObject *kwds)
   Py_RETURN_NONE;
 }
 
-/** .. method:: getexectrace() -> Optional[ExecTracer]
+/** .. method:: get_exec_trace() -> Optional[ExecTracer]
 
   Returns the currently installed :attr:`execution tracer
-  <Cursor.exectrace>`
+  <Cursor.exec_trace>`
 
   .. seealso::
 
     * :ref:`tracing`
 */
 static PyObject *
-APSWCursor_getexectrace(APSWCursor *self)
+APSWCursor_get_exec_trace(APSWCursor *self)
 {
   PyObject *ret;
 
@@ -1446,9 +1318,9 @@ APSWCursor_getexectrace(APSWCursor *self)
   return Py_NewRef(ret);
 }
 
-/** .. method:: getrowtrace() -> Optional[RowTracer]
+/** .. method:: get_row_trace() -> Optional[RowTracer]
 
-  Returns the currently installed (via :meth:`~Cursor.setrowtrace`)
+  Returns the currently installed (via :meth:`~Cursor.set_row_trace`)
   row tracer.
 
   .. seealso::
@@ -1456,7 +1328,7 @@ APSWCursor_getexectrace(APSWCursor *self)
     * :ref:`tracing`
 */
 static PyObject *
-APSWCursor_getrowtrace(APSWCursor *self)
+APSWCursor_get_row_trace(APSWCursor *self)
 {
   PyObject *ret;
   CHECK_USE(NULL);
@@ -1465,13 +1337,13 @@ APSWCursor_getrowtrace(APSWCursor *self)
   return Py_NewRef(ret);
 }
 
-/** .. method:: getconnection() -> Connection
+/** .. method:: get_connection() -> Connection
 
-  Returns the :attr:`connection` this cursor is using
+  Returns the :attr:`connection` this cursor is part of
 */
 
 static PyObject *
-APSWCursor_getconnection(APSWCursor *self)
+APSWCursor_get_connection(APSWCursor *self)
 {
   CHECK_USE(NULL);
   CHECK_CURSOR_CLOSED(NULL);
@@ -1482,7 +1354,8 @@ APSWCursor_getconnection(APSWCursor *self)
 /** .. method:: fetchall() -> list[tuple[SQLiteValue, ...]]
 
   Returns all remaining result rows as a list.  This method is defined
-  in DBAPI.  It is a longer way of doing ``list(cursor)``.
+  in DBAPI.  See :meth:`get` which does the same thing, but with the least
+  amount of structure to unpack.
 */
 static PyObject *
 APSWCursor_fetchall(APSWCursor *self)
@@ -1514,7 +1387,7 @@ APSWCursor_fetchone(APSWCursor *self)
   return res;
 }
 
-/** .. attribute:: exectrace
+/** .. attribute:: exec_trace
   :type: Optional[ExecTracer]
 
   Called with the cursor, statement and bindings for
@@ -1528,11 +1401,11 @@ APSWCursor_fetchone(APSWCursor *self)
 
     * :ref:`tracing`
     * :ref:`executiontracer`
-    * :attr:`Connection.exectrace`
+    * :attr:`Connection.exec_trace`
 
 */
 static PyObject *
-APSWCursor_get_exectrace_attr(APSWCursor *self)
+APSWCursor_get_exec_trace_attr(APSWCursor *self)
 {
   CHECK_USE(NULL);
   CHECK_CURSOR_CLOSED(NULL);
@@ -1543,14 +1416,14 @@ APSWCursor_get_exectrace_attr(APSWCursor *self)
 }
 
 static int
-APSWCursor_set_exectrace_attr(APSWCursor *self, PyObject *value)
+APSWCursor_set_exec_trace_attr(APSWCursor *self, PyObject *value)
 {
   CHECK_USE(-1);
   CHECK_CURSOR_CLOSED(-1);
 
   if (!Py_IsNone(value) && !PyCallable_Check(value))
   {
-    PyErr_Format(PyExc_TypeError, "exectrace expected a Callable");
+    PyErr_Format(PyExc_TypeError, "exec_trace expected a Callable");
     return -1;
   }
   Py_CLEAR(self->exectrace);
@@ -1559,7 +1432,7 @@ APSWCursor_set_exectrace_attr(APSWCursor *self, PyObject *value)
   return 0;
 }
 
-/** .. attribute:: rowtrace
+/** .. attribute:: row_trace
   :type: Optional[RowTracer]
 
   Called with cursor and row being returned.  You can
@@ -1573,11 +1446,11 @@ APSWCursor_set_exectrace_attr(APSWCursor *self, PyObject *value)
 
     * :ref:`tracing`
     * :ref:`rowtracer`
-    * :attr:`Connection.rowtrace`
+    * :attr:`Connection.row_trace`
 
 */
 static PyObject *
-APSWCursor_get_rowtrace_attr(APSWCursor *self)
+APSWCursor_get_row_trace_attr(APSWCursor *self)
 {
   CHECK_USE(NULL);
   CHECK_CURSOR_CLOSED(NULL);
@@ -1588,7 +1461,7 @@ APSWCursor_get_rowtrace_attr(APSWCursor *self)
 }
 
 static int
-APSWCursor_set_rowtrace_attr(APSWCursor *self, PyObject *value)
+APSWCursor_set_row_trace_attr(APSWCursor *self, PyObject *value)
 {
   CHECK_USE(-1);
   CHECK_CURSOR_CLOSED(-1);
@@ -1610,7 +1483,7 @@ APSWCursor_set_rowtrace_attr(APSWCursor *self, PyObject *value)
   :class:`Connection` this cursor is using
 */
 static PyObject *
-APSWCursor_getconnection_attr(APSWCursor *self)
+APSWCursor_get_connection_attr(APSWCursor *self)
 {
   CHECK_USE(NULL);
   CHECK_CURSOR_CLOSED(NULL);
@@ -1618,6 +1491,65 @@ APSWCursor_getconnection_attr(APSWCursor *self)
   return Py_NewRef((PyObject *)self->connection);
 }
 
+/** .. attribute:: bindings_count
+  :type: int
+
+  How many bindings are in the statement.  The ``?`` form
+  results in the largest number.  For example you could do
+  ``SELECT ?123``` in which case the count will be ``123``.
+
+  -* sqlite3_bind_parameter_count
+*/
+static PyObject *
+APSWCursor_bindings_count(APSWCursor *self)
+{
+  CHECK_USE(NULL);
+  CHECK_CURSOR_CLOSED(NULL);
+
+  return PyLong_FromLong((self->statement) ? sqlite3_bind_parameter_count(self->statement->vdbestatement) : 0);
+}
+
+/** .. attribute:: bindings_names
+  :type: tuple[str | None]
+
+  A tuple of the name of each bind parameter, or None for no name.  The
+  leading marker (``?:@$``) is omitted
+
+  .. note::
+
+    SQLite parameter numbering starts at ``1``, while Python
+    indexing starts at ``0``.
+
+  -* sqlite3_bind_parameter_name
+*/
+static PyObject *
+APSWCursor_bindings_names(APSWCursor *self)
+{
+  CHECK_USE(NULL);
+  CHECK_CURSOR_CLOSED(NULL);
+
+  int count = (self->statement) ? sqlite3_bind_parameter_count(self->statement->vdbestatement) : 0;
+
+  PyObject *res = PyTuple_New(count);
+  if (!res)
+    goto error;
+
+  for (int i = 1; i <= count; i++)
+  {
+    const char *name = sqlite3_bind_parameter_name(self->statement->vdbestatement, i);
+
+    PyObject *val = name ? PyUnicode_FromString(name + 1) : Py_NewRef(Py_None);
+    if (!val)
+      goto error;
+    PyTuple_SET_ITEM(res, i - 1, val);
+  }
+
+  return res;
+error:
+  Py_XDECREF(res);
+  return NULL;
+}
+
 /** .. attribute:: is_explain
   :type: int
 
@@ -1658,8 +1590,8 @@ APSWCursor_is_readonly(APSWCursor *self)
 /** .. attribute:: has_vdbe
   :type: bool
 
-  `True` if the SQL can be evaluated.  Comments have nothing to
-   evaluate, and so are `False`.
+  ``True`` if the SQL does anything.  Comments have nothing to
+  evaluate, and so are ``False``.
 */
 static PyObject *
 APSWCursor_has_vdbe(APSWCursor *self)
@@ -1720,6 +1652,7 @@ APSWCursor_expanded_sql(APSWCursor *self)
 
  .. list-table:: Some examples
     :header-rows: 1
+    :widths: auto
 
     * - Query
       - Result
@@ -1806,44 +1739,70 @@ error:
   return NULL;
 }
 
+static PyObject *
+APSWCursor_tp_str(APSWCursor *self)
+{
+  return PyUnicode_FromFormat("<apsw.Cursor object from %S at %p>",
+                              self->connection ? (PyObject *)self->connection : apst.closed,
+                              self);
+}
+
 static PyMethodDef APSWCursor_methods[] = {
-    {"execute", (PyCFunction)APSWCursor_execute, METH_VARARGS | METH_KEYWORDS,
+    {"execute", (PyCFunction)APSWCursor_execute, METH_FASTCALL | METH_KEYWORDS,
      Cursor_execute_DOC},
-    {"executemany", (PyCFunction)APSWCursor_executemany, METH_VARARGS | METH_KEYWORDS,
+    {"executemany", (PyCFunction)APSWCursor_executemany, METH_FASTCALL | METH_KEYWORDS,
      Cursor_executemany_DOC},
-    {"setexectrace", (PyCFunction)APSWCursor_setexectrace, METH_VARARGS | METH_KEYWORDS,
-     Cursor_setexectrace_DOC},
-    {"setrowtrace", (PyCFunction)APSWCursor_setrowtrace, METH_VARARGS | METH_KEYWORDS,
-     Cursor_setrowtrace_DOC},
-    {"getexectrace", (PyCFunction)APSWCursor_getexectrace, METH_NOARGS,
-     Cursor_getexectrace_DOC},
-    {"getrowtrace", (PyCFunction)APSWCursor_getrowtrace, METH_NOARGS,
-     Cursor_getrowtrace_DOC},
-    {"getconnection", (PyCFunction)APSWCursor_getconnection, METH_NOARGS,
-     Cursor_getconnection_DOC},
-    {"getdescription", (PyCFunction)APSWCursor_getdescription, METH_NOARGS,
-     Cursor_getdescription_DOC},
-    {"close", (PyCFunction)APSWCursor_close, METH_VARARGS | METH_KEYWORDS,
+    {"set_exec_trace", (PyCFunction)APSWCursor_set_exec_trace, METH_FASTCALL | METH_KEYWORDS,
+     Cursor_set_exec_trace_DOC},
+    {"set_row_trace", (PyCFunction)APSWCursor_set_row_trace, METH_FASTCALL | METH_KEYWORDS,
+     Cursor_set_row_trace_DOC},
+    {"get_exec_trace", (PyCFunction)APSWCursor_get_exec_trace, METH_NOARGS,
+     Cursor_get_exec_trace_DOC},
+    {"get_row_trace", (PyCFunction)APSWCursor_get_row_trace, METH_NOARGS,
+     Cursor_get_row_trace_DOC},
+    {"get_connection", (PyCFunction)APSWCursor_get_connection, METH_NOARGS,
+     Cursor_get_connection_DOC},
+    {"get_description", (PyCFunction)APSWCursor_get_description, METH_NOARGS,
+     Cursor_get_description_DOC},
+    {"close", (PyCFunction)APSWCursor_close, METH_FASTCALL | METH_KEYWORDS,
      Cursor_close_DOC},
     {"fetchall", (PyCFunction)APSWCursor_fetchall, METH_NOARGS,
      Cursor_fetchall_DOC},
     {"fetchone", (PyCFunction)APSWCursor_fetchone, METH_NOARGS,
      Cursor_fetchone_DOC},
+#ifndef APSW_OMIT_OLD_NAMES
+    {Cursor_set_exec_trace_OLDNAME, (PyCFunction)APSWCursor_set_exec_trace, METH_FASTCALL | METH_KEYWORDS,
+     Cursor_set_exec_trace_OLDDOC},
+    {Cursor_set_row_trace_OLDNAME, (PyCFunction)APSWCursor_set_row_trace, METH_FASTCALL | METH_KEYWORDS,
+     Cursor_set_row_trace_OLDDOC},
+    {Cursor_get_exec_trace_OLDNAME, (PyCFunction)APSWCursor_get_exec_trace, METH_NOARGS,
+     Cursor_get_exec_trace_OLDDOC},
+    {Cursor_get_row_trace_OLDNAME, (PyCFunction)APSWCursor_get_row_trace, METH_NOARGS,
+     Cursor_get_row_trace_OLDDOC},
+    {Cursor_get_connection_OLDNAME, (PyCFunction)APSWCursor_get_connection, METH_NOARGS,
+     Cursor_get_connection_OLDDOC},
+    {Cursor_get_description_OLDNAME, (PyCFunction)APSWCursor_get_description, METH_NOARGS,
+     Cursor_get_description_OLDDOC},
+#endif
     {0, 0, 0, 0} /* Sentinel */
 };
 
 static PyGetSetDef APSWCursor_getset[] = {
     {"description", (getter)APSWCursor_getdescription_dbapi, NULL, Cursor_description_DOC, NULL},
 #ifdef SQLITE_ENABLE_COLUMN_METADATA
-    {"description_full", (getter)APSWCursor_getdescription_full, NULL, Cursor_description_full_DOC, NULL},
+    {"description_full", (getter)APSWCursor_get_description_full, NULL, Cursor_description_full_DOC, NULL},
 #endif
     {"is_explain", (getter)APSWCursor_is_explain, NULL, Cursor_is_explain_DOC, NULL},
     {"is_readonly", (getter)APSWCursor_is_readonly, NULL, Cursor_is_readonly_DOC, NULL},
     {"has_vdbe", (getter)APSWCursor_has_vdbe, NULL, Cursor_has_vdbe_DOC, NULL},
+    {"bindings_count", (getter)APSWCursor_bindings_count, NULL, Cursor_bindings_count_DOC, NULL},
+    {"bindings_names", (getter)APSWCursor_bindings_names, NULL, Cursor_bindings_names_DOC, NULL},
     {"expanded_sql", (getter)APSWCursor_expanded_sql, NULL, Cursor_expanded_sql_DOC, NULL},
-    {"exectrace", (getter)APSWCursor_get_exectrace_attr, (setter)APSWCursor_set_exectrace_attr, Cursor_exectrace_DOC},
-    {"rowtrace", (getter)APSWCursor_get_rowtrace_attr, (setter)APSWCursor_set_rowtrace_attr, Cursor_rowtrace_DOC},
-    {"connection", (getter)APSWCursor_getconnection_attr, NULL, Cursor_connection_DOC},
+    {"exec_trace", (getter)APSWCursor_get_exec_trace_attr, (setter)APSWCursor_set_exec_trace_attr, Cursor_exec_trace_DOC},
+    {Cursor_exec_trace_OLDNAME, (getter)APSWCursor_get_exec_trace_attr, (setter)APSWCursor_set_exec_trace_attr, Cursor_exec_trace_OLDDOC},
+    {"row_trace", (getter)APSWCursor_get_row_trace_attr, (setter)APSWCursor_set_row_trace_attr, Cursor_row_trace_DOC},
+    {Cursor_row_trace_OLDNAME, (getter)APSWCursor_get_row_trace_attr, (setter)APSWCursor_set_row_trace_attr, Cursor_row_trace_OLDDOC},
+    {"connection", (getter)APSWCursor_get_connection_attr, NULL, Cursor_connection_DOC},
     {"get", (getter)APSWCursor_get, NULL, Cursor_get_DOC},
     {NULL, NULL, NULL, NULL, NULL}};
 
@@ -1862,4 +1821,5 @@ static PyTypeObject APSWCursorType = {
     .tp_getset = APSWCursor_getset,
     .tp_init = (initproc)APSWCursor_init,
     .tp_new = APSWCursor_new,
+    .tp_str = (reprfunc)APSWCursor_tp_str,
 };
