@@ -4,6 +4,14 @@
   See the accompanying LICENSE file.
 */
 
+/* msvc doesn't support vla, so do it the hard way */
+#if defined(_MSC_VER) || defined(__STDC_NO_VLA__)
+#define VLA_PYO(name, size) \
+  PyObject **name = alloca(sizeof(PyObject *) * (size))
+#else
+#define VLA_PYO(name, size) PyObject *name[size]
+#endif
+
 /* These macros are to address several issues:
 
   - Prevent simultaneous calls on the same object while the GIL is
@@ -127,7 +135,6 @@ apsw_write_unraisable(PyObject *hookobject)
 {
   assert(PyErr_Occurred());
 
-  PyObject *err_type = NULL, *err_value = NULL, *err_traceback = NULL;
   PyObject *excepthook = NULL;
   PyObject *result = NULL;
 
@@ -155,16 +162,16 @@ apsw_write_unraisable(PyObject *hookobject)
 #endif
 
   /* Get the exception details */
-  PyErr_Fetch(&err_type, &err_value, &err_traceback);
-  PyErr_NormalizeException(&err_type, &err_value, &err_traceback);
+  PY_ERR_FETCH(exc);
+  PY_ERR_NORMALIZE(exc);
 
   /* tell sqlite3_log */
-  if (err_value && 0 == Py_EnterRecursiveCall("apsw_write_unraisable forwarding to sqlite3_log"))
+  if (exc && 0 == Py_EnterRecursiveCall("apsw_write_unraisable forwarding to sqlite3_log"))
   {
-    PyObject *message = PyObject_Str(err_value);
+    PyObject *message = PyObject_Str(exc);
     const char *utf8 = message ? PyUnicode_AsUTF8(message) : "failed to get string of error";
     PyErr_Clear();
-    sqlite3_log(SQLITE_ERROR, "apsw_write_unraisable %s: %s", Py_TypeName(err_value), utf8);
+    sqlite3_log(SQLITE_ERROR, "apsw_write_unraisable %s: %s", Py_TYPE(exc)->tp_name, utf8);
     Py_CLEAR(message);
     Py_LeaveRecursiveCall();
   }
@@ -173,11 +180,16 @@ apsw_write_unraisable(PyObject *hookobject)
 
   if (hookobject)
   {
-    excepthook = PyObject_GetAttrString(hookobject, "excepthook");
+    excepthook = PyObject_GetAttr(hookobject, apst.excepthook);
     PyErr_Clear();
     if (excepthook)
     {
-      result = PyObject_CallFunction(excepthook, "(OOO)", OBJ(err_type), OBJ(err_value), OBJ(err_traceback));
+#if PY_VERSION_HEX < 0x030c0000
+      PyObject *vargs[] = {NULL, OBJ(exctype), OBJ(exc), OBJ(exctraceback)};
+#else
+      PyObject *vargs[] = {NULL, (PyObject *)Py_TYPE(OBJ(exc)), OBJ(exc), Py_None};
+#endif
+      result = PyObject_Vectorcall(excepthook, vargs + 1, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
       if (result)
         goto finally;
     }
@@ -192,10 +204,17 @@ apsw_write_unraisable(PyObject *hookobject)
     PyObject *arg = PyStructSequence_New(&apsw_unraisable_info_type);
     if (arg)
     {
-      PyStructSequence_SetItem(arg, 0, Py_NewRef(OBJ(err_type)));
-      PyStructSequence_SetItem(arg, 1, Py_NewRef(OBJ(err_value)));
-      PyStructSequence_SetItem(arg, 2, Py_NewRef(OBJ(err_traceback)));
-      result = PyObject_CallFunction(excepthook, "(N)", arg);
+#if PY_VERSION_HEX < 0x030c0000
+      PyStructSequence_SetItem(arg, 0, Py_NewRef(OBJ(exctype)));
+      PyStructSequence_SetItem(arg, 1, Py_NewRef(OBJ(exc)));
+      PyStructSequence_SetItem(arg, 2, Py_NewRef(OBJ(exctraceback)));
+#else
+      PyStructSequence_SetItem(arg, 0, Py_NewRef((PyObject *)Py_TYPE(OBJ(exc))));
+      PyStructSequence_SetItem(arg, 1, Py_NewRef(exc));
+#endif
+      PyObject *vargs[] = {NULL, arg};
+      result = PyObject_Vectorcall(excepthook, vargs + 1, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
+      Py_DECREF(arg);
       if (result)
         goto finally;
     }
@@ -207,7 +226,13 @@ apsw_write_unraisable(PyObject *hookobject)
   {
     Py_INCREF(excepthook); /* borrowed reference from PySys_GetObject so we increment */
     PyErr_Clear();
-    result = PyObject_CallFunction(excepthook, "(OOO)", OBJ(err_type), OBJ(err_value), OBJ(err_traceback));
+#if PY_VERSION_HEX < 0x030c0000
+    PyObject *vargs[] = {NULL, OBJ(exctype), OBJ(exc), OBJ(exctraceback)};
+#else
+    PyObject *vargs[] = {NULL, (PyObject *)Py_TYPE(OBJ(exc)), OBJ(exc), Py_None};
+#endif
+
+    result = PyObject_Vectorcall(excepthook, vargs + 1, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
     if (result)
       goto finally;
   }
@@ -215,14 +240,16 @@ apsw_write_unraisable(PyObject *hookobject)
   /* remove any error from callback failure since we'd have to call
      ourselves to raise it! */
   PyErr_Clear();
-  PyErr_Display(err_type, err_value, err_traceback);
+#if PY_VERSION_HEX < 0x030c0000
+  PyErr_Display(exctype, exc, exctraceback);
+#else
+  PyErr_DisplayException(exc);
+#endif
 
 finally:
   Py_XDECREF(excepthook);
   Py_XDECREF(result);
-  Py_XDECREF(err_traceback);
-  Py_XDECREF(err_value);
-  Py_XDECREF(err_type);
+  PY_ERR_CLEAR(exc);
   PyErr_Clear(); /* being paranoid - make sure no errors on return */
 }
 
